@@ -11,10 +11,9 @@
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
 import socket
-from tests import unittest
 
-from mock import Mock, patch, sentinel
-from botocore.vendored.requests import ConnectionError
+from tests import mock
+from tests import unittest
 
 from botocore.compat import six
 from botocore.awsrequest import AWSRequest
@@ -22,12 +21,14 @@ from botocore.endpoint import Endpoint, DEFAULT_TIMEOUT
 from botocore.endpoint import EndpointCreator
 from botocore.exceptions import EndpointConnectionError
 from botocore.exceptions import ConnectionClosedError
+from botocore.exceptions import HTTPClientError
 from botocore.httpsession import URLLib3Session
 from botocore.model import OperationModel, ServiceId
+from botocore.model import ServiceModel, StructureShape
 
 
-def request_dict():
-    return {
+def request_dict(**kwargs):
+    base = {
         'headers': {},
         'body': '',
         'url_path': '/',
@@ -36,6 +37,8 @@ def request_dict():
         'url': 'https://example.com',
         'context': {}
     }
+    base.update(kwargs)
+    return base
 
 
 class RecordStreamResets(six.StringIO):
@@ -51,21 +54,21 @@ class RecordStreamResets(six.StringIO):
 class TestEndpointBase(unittest.TestCase):
 
     def setUp(self):
-        self.op = Mock()
+        self.op = mock.Mock()
         self.op.has_streaming_output = False
         self.op.has_event_stream_output = False
         self.op.metadata = {'protocol': 'json'}
-        self.event_emitter = Mock()
+        self.event_emitter = mock.Mock()
         self.event_emitter.emit.return_value = []
-        self.factory_patch = patch(
+        self.factory_patch = mock.patch(
             'botocore.parsers.ResponseParserFactory')
         self.factory = self.factory_patch.start()
         self.endpoint = Endpoint(
             'https://ec2.us-west-2.amazonaws.com/',
             endpoint_prefix='ec2',
             event_emitter=self.event_emitter)
-        self.http_session = Mock()
-        self.http_session.send.return_value = Mock(
+        self.http_session = mock.Mock()
+        self.http_session.send.return_value = mock.Mock(
             status_code=200, headers={}, content=b'{"Foo": "bar"}',
         )
         self.endpoint.http_session = self.http_session
@@ -73,6 +76,27 @@ class TestEndpointBase(unittest.TestCase):
     def tearDown(self):
         self.factory_patch.stop()
 
+    def get_emitter_responses(self, num_retries=0, sleep_time=0):
+        emitter_responses = []
+        response_request_emitter_responses = [
+            [(None, None)],  # emit() response for request-created
+            [(None, None)],  # emit() response for before-send
+            [(None, None)],  # emit() response for response-received
+        ]
+        for _ in range(num_retries):
+            emitter_responses.extend(response_request_emitter_responses)
+            # emit() response for retry for sleep time
+            emitter_responses.append([(None, sleep_time)])
+        emitter_responses.extend(response_request_emitter_responses)
+        # emit() response for no retry
+        emitter_responses.append([(None, None)])
+        return emitter_responses
+
+    def get_events_emitted(self, event_emitter):
+        return [
+            call_arg[0][0] for call_arg in event_emitter.emit.call_args_list
+        ]
+
 
 class TestEndpointFeatures(TestEndpointBase):
 
@@ -96,120 +120,118 @@ class TestEndpointFeatures(TestEndpointBase):
     def test_make_request_with_context(self):
         r = request_dict()
         r['context'] = {'signing': {'region': 'us-west-2'}}
-        with patch('botocore.endpoint.Endpoint.prepare_request') as prepare:
+        with mock.patch('botocore.endpoint.Endpoint.prepare_request') as prepare:
             self.endpoint.make_request(self.op, r)
         request = prepare.call_args[0][0]
         self.assertEqual(request.context['signing']['region'], 'us-west-2')
 
+    def test_parses_modeled_exception_fields(self):
+        # Setup the service model to have exceptions to generate the mapping
+        self.service_model = mock.Mock(spec=ServiceModel)
+        self.op.service_model = self.service_model
+        self.exception_shape = mock.Mock(spec=StructureShape)
+        shape_for_error_code = self.service_model.shape_for_error_code
+        shape_for_error_code.return_value = self.exception_shape
+
+        r = request_dict()
+        self.http_session.send.return_value = mock.Mock(
+            status_code=400, headers={}, content=b'',
+        )
+        parser = mock.Mock()
+        parser.parse.side_effect = [
+            {
+                'Error': {
+                    'Code': 'ExceptionShape',
+                    'Message': 'Some message',
+                }
+            },
+            {'SomeField': 'Foo'},
+        ]
+        self.factory.return_value.create_parser.return_value = parser
+        _, response = self.endpoint.make_request(self.op, r)
+        # The parser should be called twice, once for the original
+        # error parse and once again for the modeled exception parse
+        self.assertEqual(parser.parse.call_count, 2)
+        parse_calls = parser.parse.call_args_list
+        self.assertEqual(parse_calls[1][0][1], self.exception_shape)
+        self.assertEqual(parse_calls[0][0][1], self.op.output_shape)
+        expected_response = {
+            'Error': {
+                'Code': 'ExceptionShape',
+                'Message': 'Some message',
+            },
+            'SomeField': 'Foo',
+        }
+        self.assertEqual(response, expected_response)
+
+
 class TestRetryInterface(TestEndpointBase):
     def setUp(self):
         super(TestRetryInterface, self).setUp()
         self.retried_on_exception = None
-        self._operation = Mock(spec=OperationModel)
-        self._operation.service_model.service_id = ServiceId('ec2')
+        self._operation = mock.Mock(spec=OperationModel)
+        self._operation.name = 'DescribeInstances'
+        self._operation.metadata = {'protocol': 'query'}
+        self._operation.service_model.service_id = ServiceId('EC2')
+        self._operation.has_streaming_output = False
+        self._operation.has_event_stream_output = False
+
+    def assert_events_emitted(self, event_emitter, expected_events):
+        self.assertEqual(
+            self.get_events_emitted(event_emitter), expected_events)
 
     def test_retry_events_are_emitted(self):
-        op = self._operation
-        op.name = 'DescribeInstances'
-        op.metadata = {'protocol': 'query'}
-        op.has_streaming_output = False
-        op.has_event_stream_output = False
-        self.endpoint.make_request(op, request_dict())
+        self.endpoint.make_request(self._operation, request_dict())
         call_args = self.event_emitter.emit.call_args
         self.assertEqual(call_args[0][0],
                          'needs-retry.ec2.DescribeInstances')
 
     def test_retry_events_can_alter_behavior(self):
-        op = self._operation
-        op.name = 'DescribeInstances'
-        op.metadata = {'protocol': 'json'}
-        op.has_event_stream_output = False
-        self.event_emitter.emit.side_effect = [
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent.
-            [(None, 0)],       # Check if retry needed. Retry needed.
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent.
-            [(None, None)]     # Check if retry needed. Retry not needed.
-        ]
-        self.endpoint.make_request(op, request_dict())
-        call_args = self.event_emitter.emit.call_args_list
-        self.assertEqual(self.event_emitter.emit.call_count, 6)
-        # Check that all of the events are as expected.
-        self.assertEqual(call_args[0][0][0],
-                         'request-created.ec2.DescribeInstances')
-        self.assertEqual(call_args[1][0][0],
-                         'before-send.ec2.DescribeInstances')
-        self.assertEqual(call_args[2][0][0],
-                         'needs-retry.ec2.DescribeInstances')
-        self.assertEqual(call_args[3][0][0],
-                         'request-created.ec2.DescribeInstances')
-        self.assertEqual(call_args[4][0][0],
-                         'before-send.ec2.DescribeInstances')
-        self.assertEqual(call_args[5][0][0],
-                         'needs-retry.ec2.DescribeInstances')
+        self.event_emitter.emit.side_effect = self.get_emitter_responses(
+            num_retries=1)
+        self.endpoint.make_request(self._operation, request_dict())
+        self.assert_events_emitted(
+            self.event_emitter,
+            expected_events=[
+                'request-created.ec2.DescribeInstances',
+                'before-send.ec2.DescribeInstances',
+                'response-received.ec2.DescribeInstances',
+                'needs-retry.ec2.DescribeInstances',
+            ] * 2
+        )
 
     def test_retry_on_socket_errors(self):
-        op = self._operation
-        op.name = 'DescribeInstances'
-        op.has_event_stream_output = False
-        self.event_emitter.emit.side_effect = [
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent.
-            [(None, 0)],       # Check if retry needed. Retry needed.
-            [(None, None)],    # Request created
-            [(None, None)],    # Request sent.
-            [(None, None)]     # Check if retry needed. Retry not needed.
-        ]
-        self.http_session.send.side_effect = ConnectionError()
-        with self.assertRaises(ConnectionError):
-            self.endpoint.make_request(op, request_dict())
-        call_args = self.event_emitter.emit.call_args_list
-        self.assertEqual(self.event_emitter.emit.call_count, 6)
-        # Check that all of the events are as expected.
-        self.assertEqual(call_args[0][0][0],
-                         'request-created.ec2.DescribeInstances')
-        self.assertEqual(call_args[1][0][0],
-                         'before-send.ec2.DescribeInstances')
-        self.assertEqual(call_args[2][0][0],
-                         'needs-retry.ec2.DescribeInstances')
-        self.assertEqual(call_args[3][0][0],
-                         'request-created.ec2.DescribeInstances')
-        self.assertEqual(call_args[4][0][0],
-                         'before-send.ec2.DescribeInstances')
-        self.assertEqual(call_args[5][0][0],
-                         'needs-retry.ec2.DescribeInstances')
+        self.event_emitter.emit.side_effect = self.get_emitter_responses(
+            num_retries=1)
+        self.http_session.send.side_effect = HTTPClientError(error='wrapped')
+        with self.assertRaises(HTTPClientError):
+            self.endpoint.make_request(self._operation, request_dict())
+        self.assert_events_emitted(
+            self.event_emitter,
+            expected_events=[
+                'request-created.ec2.DescribeInstances',
+                'before-send.ec2.DescribeInstances',
+                'response-received.ec2.DescribeInstances',
+                'needs-retry.ec2.DescribeInstances',
+            ] * 2
+        )
 
     def test_retry_attempts_added_to_response_metadata(self):
-        op = Mock(name='DescribeInstances')
-        op.metadata = {'protocol': 'query'}
-        op.has_event_stream_output = False
-        self.event_emitter.emit.side_effect = [
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent
-            [(None, 0)],       # Check if retry needed. Retry needed.
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent
-            [(None, None)]     # Check if retry needed. Retry not needed.
-        ]
-        parser = Mock()
+        self.event_emitter.emit.side_effect = self.get_emitter_responses(
+            num_retries=1)
+        parser = mock.Mock()
         parser.parse.return_value = {'ResponseMetadata': {}}
         self.factory.return_value.create_parser.return_value = parser
-        response = self.endpoint.make_request(op, request_dict())
+        response = self.endpoint.make_request(self._operation, request_dict())
         self.assertEqual(response[1]['ResponseMetadata']['RetryAttempts'], 1)
 
     def test_retry_attempts_is_zero_when_not_retried(self):
-        op = Mock(name='DescribeInstances', metadata={'protocol': 'query'})
-        op.has_event_stream_output = False
-        self.event_emitter.emit.side_effect = [
-            [(None, None)],    # Request created.
-            [(None, None)],    # Request sent.
-            [(None, None)],    # Check if retry needed. Retry needed.
-        ]
-        parser = Mock()
+        self.event_emitter.emit.side_effect = self.get_emitter_responses(
+            num_retries=0)
+        parser = mock.Mock()
         parser.parse.return_value = {'ResponseMetadata': {}}
         self.factory.return_value.create_parser.return_value = parser
-        response = self.endpoint.make_request(op, request_dict())
+        response = self.endpoint.make_request(self._operation, request_dict())
         self.assertEqual(response[1]['ResponseMetadata']['RetryAttempts'], 0)
 
 
@@ -228,7 +250,7 @@ class TestS3ResetStreamOnRetry(TestEndpointBase):
             return 0
 
     def test_reset_stream_on_retry(self):
-        op = Mock()
+        op = mock.Mock()
         body = RecordStreamResets('foobar')
         op.name = 'PutObject'
         op.has_streaming_output = True
@@ -236,17 +258,9 @@ class TestS3ResetStreamOnRetry(TestEndpointBase):
         op.metadata = {'protocol': 'rest-xml'}
         request = request_dict()
         request['body'] = body
-        self.event_emitter.emit.side_effect = [
-            [(None, None)],   # Request created.
-            [(None, None)],   # Request sent.
-            [(None, 0)],      # Check if retry needed. Needs Retry.
-            [(None, None)],   # Request created.
-            [(None, None)],   # Request sent.
-            [(None, 0)],      # Check if retry needed again. Needs Retry.
-            [(None, None)],   # Request created.
-            [(None, None)],   # Request sent.
-            [(None, None)],   # Finally emit no rety is needed.
-        ]
+        self.event_emitter.emit.side_effect = self.get_emitter_responses(
+            num_retries=2
+        )
         self.endpoint.make_request(op, request)
         # 2 seeks for the resets and 6 (2 per creation) for content-length
         self.assertEqual(body.total_resets, 8)
@@ -264,14 +278,14 @@ class TestEventStreamBody(TestEndpointBase):
 
 class TestEndpointCreator(unittest.TestCase):
     def setUp(self):
-        self.service_model = Mock(
+        self.service_model = mock.Mock(
             endpoint_prefix='ec2', signature_version='v2',
             signing_name='ec2')
         self.environ = {}
-        self.environ_patch = patch('os.environ', self.environ)
+        self.environ_patch = mock.patch('os.environ', self.environ)
         self.environ_patch.start()
-        self.creator = EndpointCreator(Mock())
-        self.mock_session = Mock(spec=URLLib3Session)
+        self.creator = EndpointCreator(mock.Mock())
+        self.mock_session = mock.Mock(spec=URLLib3Session)
 
     def tearDown(self):
         self.environ_patch.stop()
